// Reads internal Temperature Sensor with very high resolution, readings send to USB.For ATmega328 types. R.Schouten, www.RS-ELC.nl
// Detects small drops in temperature, examples: can detect blowing on the chip from 30cm distance or a fast movement of the board
// Concept of airflow detector: Arduino-Chip heats up inside and an increased airflow at the surface cools it down a bit
// To have it working, after power-up wait until chip temperature settles (check readings via USB) after a few minutes
// This readout is high precision (fluctuates <0.02 celsius) but is not accurate (means the absolute value can be several degrees off) 
// For applications detecting a small change you only need high precision so this works well here.
 
// datasheet information: tempsensor typ gives 1mV/celsius and reads 314mV for 25celsius
// 25celsius is then ADCvalue 353 on Vref=1.1V
// 1mV/celsius is 0.93LSB/celsius on Vref=1.1V
// use 6400x averaging to increase resolution by 80x (dithering), adds approx +6bits, results in readings 0.01 celsius resolution
// note: assumes chip temperatures above zero celsius 
// note: works best on SMD chip version. DIL-chip version is larger, giving slower thermal response, and more isolated I presume
// note: This is a very senstive temperature meter,readout flucuations can be <0.02 celsius but any airflow will increase this. 

unsigned int wADCread;// ADC single readout value
unsigned long ADCsum;// data register to hold summation of ADC-readings to make one temp readout result
unsigned long LAST_ADCsum;// previous temp readout result, difference is used to detect airflow
unsigned int avgcnt=0;// counter to track number of ADC-readings taken
float t;// temperature value result that is send to PC
unsigned int TempOffset=320;//320 gives temp offset to start scale at zero celsius (not a gain calibration, only shifts scale)
//optional: check offset for individual chip by reading cold-boot=ambient and then correct TempOffset to get the right value
//this calibration is optional for getting roughly correct absolute values, not needed for just the airflowdetection to work.  

void setup()
{
  Serial.begin(9600);
  Serial.println(F("Internal Temperature Sensor Readout"));
  // The internal temperature has to be read with the ADC reference set to 1.1V to get highest resolution.
  // ADC Channel 8 is temp sensor
  // Set the internal reference and ADC mux to sensor
  ADMUX = (_BV(REFS1) | _BV(REFS0) | _BV(MUX3));
  ADCSRA |= _BV(ADEN);  // enable the ADC
}

void loop()
{
 //clear registers at the start of each measurement:
 avgcnt=0;//clear the counter for the number of ADC-readings taken
 ADCsum=0;//clear the summation register that holds the ADC-readings for one averaged temp measurement

 //start collecting 6400 readings for one temp measurement:------------------------------------------------
 while(avgcnt < 6400){// do 6400 ADC-readings to increase resolution of one temperature measurement
   ADCSRA |= _BV(ADSC);  // Start the ADC
   while (bit_is_set(ADCSRA,ADSC));// Detect end-of-conversion
   wADCread = ADCW; // Reading register "ADCW" takes care of how to read ADCL and ADCH.
   ADCsum=ADCsum+(wADCread-TempOffset);//subtract the temperature offset calibration value
   avgcnt++;//do next ADC reading
 }
 //finished collecting readings for one temp measurement ---------------------------------------------

 t = (ADCsum/6.88);// / scale result, gain calibration to millicelsius

 Serial.print(t/1000);// send measurement(in celsius) to computer, divided by 1000 to get celsius
 Serial.println(" celsius");

 //switch off build-in LED pin13 if temperature goes down compared to previous measurement
 //detects airflow by for example blowing over board, 250 is treshold value. Higher value makes less sensitive
 if (ADCsum > (LAST_ADCsum-250)){digitalWrite(13,HIGH);}// if temperature does not go down fast, switch on LED
 else {digitalWrite(13,LOW);}//else: sudden cooldown detected, switch off LED
 LAST_ADCsum=ADCsum;// remember latest reading for next compare
}
